import asyncio

import pytest
from httpx import ASGITransport, AsyncClient
from telicent_labels import TelicentModel

from telicent_lbapi.context import ModelContext
from telicent_lbapi.rest_service import create_app

from .helpers.custom_endpoints import AnotherModel, MyModel, custom_router

app, config = create_app(model_class=TelicentModel, custom_router=custom_router)


# @pytest.fixture(scope="module")
# def anyio_backend():
#     return 'asyncio'

@pytest.mark.asyncio
async def test_concurrent_requests(anyio_backend, monkeypatch):
    # track the model classes used
    model_classes = []

    def mock_set_model_class(model):
        model_classes.append(model)
        original_set_model_class(model)

    #  prevent recursion depth errors
    original_set_model_class = ModelContext.set_model_class
    monkeypatch.setattr(ModelContext, "set_model_class", mock_set_model_class)
    transport = ASGITransport(app=app)
    async with AsyncClient(transport=transport, base_url="http://test") as ac:
        tasks = []

        for _ in range(100):
            payload = {
                "identifier": "ItemA",
                "classification": "S",
                "permittedOrgs": ["ABC", "DEF", "HIJ"],
                "permittedNats": ["GBR", "FRA", "IRL"],
                "orGroups": ["Apple", "SOMETHING"],
                "andGroups": ["doctor", "admin"],
                "originator": "TestOriginator",
                "custodian": "TestCustodian",
                "policyRef": "TestPolicyRef",
                "dataSet": ["ds1", "ds2"],
                "authRef": ["ref1", "ref2"],
                "dispositionDate": "2023-01-01T23:59:59.003426-05:00",
                "dispositionProcess": "disp-process-1",
                "dissemination": ["news", "articles"]
            }
            tasks.append(ac.post("/api/v1/ingest", json=payload))

        for _ in range(100):
            tasks.append(ac.get("/custom/custom-endpoint-mymodel"))

        for _ in range(100):
            tasks.append(ac.get("/custom/custom-endpoint-anothermodel"))

        responses = await asyncio.gather(*tasks)

        # responses for /ingest endpoint
        for response in responses[:100]:
            assert response.status_code == 200
            assert response.json()["status"] == "success"

        # responses for /custom-endpoint-mymodel
        for response in responses[100:200]:
            assert response.status_code == 200
            assert response.json()["message"] == "This is a custom endpoint using MyModel"

        # responses for /custom-endpoint-anothermodel
        for response in responses[200:]:
            assert response.status_code == 200
            assert response.json()["message"] == f"This is a custom endpoint using {AnotherModel.__name__}"

    # count how many times context was set for each model
    ingest_count = sum(1 for model in model_classes if model == TelicentModel)
    mymodel_count = sum(1 for model in model_classes if model == MyModel)
    anothermodel_count = sum(1 for model in model_classes if model == AnotherModel)

    assert ingest_count == 300  # 300 cause middleware sets it after every response
    assert mymodel_count == 100
    assert anothermodel_count == 100



import pytest

from telicent_lbapi.utils.cache import clear_cache


@pytest.fixture(autouse=True)
def run_around_tests():
    clear_cache()
    yield
    clear_cache()


@pytest.mark.asyncio
async def test_ingest_endpoint(async_client):
    payload = {
        "identifier": "ItemA",
        "classification": "S",
        "permittedOrgs": ["ABC", "DEF", "HIJ"],
        "permittedNats": ["GBR", "FRA", "IRL"],
        "orGroups": ["Apple", "SOMETHING"],
        "andGroups": ["doctor", "admin"],
        "originator": "TestOriginator",
        "custodian": "TestCustodian",
        "policyRef": "TestPolicyRef",
        "dataSet": ["ds1", "ds2"],
        "authRef": ["ref1", "ref2"],
        "dispositionDate": "2023-01-01T23:59:59.003426-05:00",
        "dispositionProcess": "disp-process-1",
        "dissemination": ["news", "articles"]
    }

    response = await async_client.post("/api/v1/ingest", json=payload)
    assert response.status_code == 200
    assert response.json()["status"] == "success"


@pytest.mark.asyncio
async def test_custom_endpoint_mymodel(async_client):
    response = await async_client.get("/custom/custom-endpoint-mymodel")
    assert response.status_code == 200
    assert response.json()["message"] == "This is a custom endpoint using MyModel"


@pytest.mark.asyncio
async def test_custom_endpoint_anothermodel(async_client):
    response = await async_client.get("/custom/custom-endpoint-anothermodel")
    assert response.status_code == 200
    assert response.json()["message"] == "This is a custom endpoint using AnotherModel"


@pytest.mark.asyncio
async def test_ingest_invalid_payload(async_client):
    payload = {"invalid": "data"}

    response = await async_client.post("/api/v1/ingest", json=payload)
    assert response.status_code == 400
    assert response.json()["detail"] == "Validation error occurred."



from datetime import datetime, timedelta

import pytest

from telicent_lbapi.utils.cache import cache_store, clear_cache


@pytest.fixture(autouse=True)
def run_around_tests():
    clear_cache()
    yield
    clear_cache()


@pytest.mark.asyncio
async def test_cache_hits(async_client):
    payload = {
        "identifier": "ItemA",
        "classification": "S",
        "permittedOrgs": ["ABC", "DEF", "HIJ"],
        "permittedNats": ["GBR", "FRA", "IRL"],
        "orGroups": ["Apple", "SOMETHING"],
        "andGroups": ["doctor", "admin"],
        "originator": "TestOriginator",
        "custodian": "TestCustodian",
        "policyRef": "TestPolicyRef",
        "dataSet": ["ds1", "ds2"],
        "authRef": ["ref1", "ref2"],
        "dispositionDate": "2023-01-01T23:59:59.003426-05:00",
        "dispositionProcess": "disp-process-1",
        "dissemination": ["news", "articles"]
    }

    response = await async_client.post("/api/v1/ingest", json=payload)
    assert response.status_code == 200
    assert response.json()["status"] == "success"

    assert len(cache_store) == 1

    response = await async_client.post("/api/v1/ingest", json=payload)
    assert response.status_code == 200
    assert response.json()["status"] == "success"

    assert len(cache_store) == 1


@pytest.mark.asyncio
async def test_cache_expiration(async_client):
    payload = {
        "identifier": "ItemA",
        "classification": "S",
        "permittedOrgs": ["ABC", "DEF", "HIJ"],
        "permittedNats": ["GBR", "FRA", "IRL"],
        "orGroups": ["Apple", "SOMETHING"],
        "andGroups": ["doctor", "admin"],
        "originator": "TestOriginator",
        "custodian": "TestCustodian",
        "policyRef": "TestPolicyRef",
        "dataSet": ["ds1", "ds2"],
        "authRef": ["ref1", "ref2"],
        "dispositionDate": "2023-01-01T23:59:59.003426-05:00",
        "dispositionProcess": "disp-process-1",
        "dissemination": ["news", "articles"]
    }

    response = await async_client.post("/api/v1/ingest", json=payload)
    assert response.status_code == 200
    assert response.json()["status"] == "success"

    assert len(cache_store) == 1

    for key in cache_store:
        cache_store[key] = (cache_store[key][0], datetime.now() - timedelta(seconds=1))

    response = await async_client.post("/api/v1/ingest", json=payload)
    assert response.status_code == 200
    assert response.json()["status"] == "success"
    assert len(cache_store) == 1


@pytest.mark.asyncio
async def test_cache_miss(async_client):
    payload_1 = {
        "identifier": "ItemA",
        "classification": "S",
        "permittedOrgs": ["ABC", "DEF", "HIJ"],
        "permittedNats": ["GBR", "FRA", "IRL"],
        "orGroups": ["Apple", "SOMETHING"],
        "andGroups": ["doctor", "admin"],
        "originator": "TestOriginator",
        "custodian": "TestCustodian",
        "policyRef": "TestPolicyRef",
        "dataSet": ["ds1", "ds2"],
        "authRef": ["ref1", "ref2"],
        "dispositionDate": "2023-01-01T23:59:59.003426-05:00",
        "dispositionProcess": "disp-process-1",
        "dissemination": ["news", "articles"]
    }

    payload_2 = {
        "identifier": "ItemB",
        "classification": "S",
        "permittedOrgs": ["XYZ"],
        "permittedNats": ["USA"],
        "orGroups": ["Orange"],
        "andGroups": ["teacher"],
        "originator": "TestOriginatorB",
        "custodian": "TestCustodianB",
        "policyRef": "TestPolicyRefB",
        "dataSet": ["ds3"],
        "authRef": ["ref3"],
        "dispositionDate": "2023-01-01T23:59:59.003426-05:00",
        "dispositionProcess": "disp-process-2",
        "dissemination": ["blogs"]
    }

    response = await async_client.post("/api/v1/ingest", json=payload_1)
    assert response.status_code == 200
    assert response.json()["status"] == "success"

    assert len(cache_store) == 1

    response = await async_client.post("/api/v1/ingest", json=payload_2)
    assert response.status_code == 200
    assert response.json()["status"] == "success"

    assert len(cache_store) == 2



import asyncio
import hashlib
import json
import logging
from datetime import datetime, timedelta

from cachetools import LRUCache
from pydantic import ValidationError

from telicent_lbapi.core.api_config.config import load_config
from telicent_lbapi.services.label_builder_service import async_build_security_labels

__license__ = """
Copyright (c) Telicent Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

log = logging.getLogger(__name__)

conf = load_config()
CACHE_SIZE: int = int(conf.cache_size)
CACHE_TIMEOUT: int = conf.cache_timeout
cache_store: LRUCache = LRUCache(maxsize=CACHE_SIZE)
cache_lock: asyncio.Lock = asyncio.Lock()


def clear_cache():
    global cache_store
    cache_store.clear()


def hash_dict(data: dict) -> str:
    sorted_data_str = json.dumps(data, sort_keys=True)
    return hashlib.sha256(sorted_data_str.encode()).hexdigest()


async def get_cached_labels(header_data: dict):
    cache_key = hash_dict(header_data)

    async with cache_lock:
        if cache_key in cache_store:
            cached_value, expiry_time = cache_store[cache_key]
            if datetime.now() < expiry_time:
                logging.debug(f"Cache entry found. {cache_key=}, {expiry_time=}")
                return cached_value

        try:
            security_labels = await async_build_security_labels(data_header=header_data)
        except ValidationError as e:
            log.error("Validation error occurred while building security labels.", exc_info=e)
            raise
        except Exception as e:
            log.error("An error occurred while building security labels.", exc_info=e)
            raise

        cache_data = (security_labels, datetime.now() + timedelta(seconds=CACHE_TIMEOUT))
        log.debug(f"Adding new key to cache={cache_key}, relative expiry={cache_data[1]}")

        cache_store[cache_key] = cache_data

    return security_labels



import logging

from starlette.types import ASGIApp, Receive, Scope, Send

from telicent_lbapi.context import ModelContext

__license__ = """
Copyright (c) Telicent Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

log = logging.getLogger(__name__)


class ModelContextMiddleware:
    def __init__(self, app: ASGIApp, default_model_class: type):
        self.app = app
        self.default_model_class = default_model_class

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] == "http":
            async def send_wrapper(message):
                if message["type"] == "http.response.start":
                    ModelContext.set_model_class(self.default_model_class)
                    log.debug(f"Model class reverted to default in middleware: "
                              f"{ModelContext.get_model_class().__name__}")
                await send(message)

            await self.app(scope, receive, send_wrapper)
        else:
            await self.app(scope, receive, send)



import logging
import os
from logging.handlers import RotatingFileHandler

from pythonjsonlogger import jsonlogger
from starlette_context import context
from starlette_context.errors import ContextDoesNotExistError

from telicent_lbapi.context import ModelContext
from telicent_lbapi.core.api_config.config import ApiConfig

__license__ = """
Copyright (c) Telicent Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class ModelClassFilter(logging.Filter):
    def filter(self, record):
        try:
            model_class = ModelContext.get_model_class()
            record.model_class = model_class.__name__
        except AttributeError:
            record.model_class = 'N/A'
        except LookupError:
            record.model_class = 'UNKNOWN'
        return True


class RequestIdFilter(logging.Filter):
    def filter(self, record):
        try:
            record.request_id = context.data.get("X-Request-ID", "unknown")
        except ContextDoesNotExistError:
            record.request_id = "N/A"

        return True


def configure_logging(config: ApiConfig):
    log_to_console = config.log_to_console
    log_level = logging.getLevelName(config.logging_level)
    log_dir = config.log_dir
    log_filename = config.log_filename

    log_file = log_dir / log_filename
    log_file.parent.mkdir(parents=True, exist_ok=True)

    if os.getenv('LABEL_BUILDER_SERVICE_DEBUG', False):
        log_level = logging.DEBUG

    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)

    request_id_filter = RequestIdFilter()
    model_class_filter = ModelClassFilter()

    console_handler = logging.StreamHandler()

    if log_to_console:
        console_handler.setLevel(log_level)
        console_handler.addFilter(request_id_filter)
        console_handler.addFilter(model_class_filter)
        console_handler.setFormatter(logging.Formatter(
            "%(asctime)s [%(levelname)s] [%(name)s:%(funcName)s:%(lineno)d]"
            " [Request ID: %(request_id)s]"
            " [Model Class: %(model_class)s] - %(message)s"
        ))
        root_logger.addHandler(console_handler)

    file_handler = RotatingFileHandler(
        filename=str(log_file), maxBytes=1024 * 1024 * 4, backupCount=3
    )
    file_handler.setLevel(log_level)
    file_handler.addFilter(request_id_filter)
    file_handler.addFilter(model_class_filter)

    keys_to_log = [
        'asctime',
        'created',
        'levelname',
        'filename',
        'funcName',
        'lineno',
        'message',
        'processName',
        'threadName',
        'request_id',
        'model_class'
    ]

    def log_format(keys):
        return [f'%({i:s})s' for i in keys]

    custom_format = ' '.join(log_format(keys_to_log))
    json_formatter = jsonlogger.JsonFormatter(custom_format)
    file_handler.setFormatter(json_formatter)
    root_logger.addHandler(file_handler)



import os
import platform
from dataclasses import dataclass
from pathlib import Path

from dotenv import load_dotenv
from telicent_lib.config import Configurator

from telicent_lbapi.utils.helpers import str_to_bool

__license__ = """
Copyright (c) Telicent Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

log_filename_default = 'label-builder-service.log'


def default_root_dir():
    system = platform.system()
    if system in ['Linux', 'Darwin']:
        log_root_dir_default = Path('/tmp/logs')
    elif system == 'Windows':
        log_root_dir_default = Path(os.getenv('TEMP')) / 'logs'
    else:
        raise OSError("Unsupported operating system. Try setting log directory manually.")

    return log_root_dir_default


@dataclass
class ApiConfig:
    logging_level: str
    cache_size: int
    cache_timeout: int
    server_port: int
    server_host: str
    log_to_console: bool
    log_dir: Path
    log_filename: str


def load_config() -> ApiConfig:
    load_dotenv()
    config = Configurator()
    logging_level = config.get(
        "LOGGING_LEVEL", "INFO"
    )
    cache_size = config.get(
        "CACHE_SIZE", default=1000000
    )
    cache_timeout = config.get(
        "CACHE_TIMEOUT", default=300
    )
    server_port = config.get(
        "RESTAPI_PORT", default=8000
    )
    server_host = config.get(
        "RESTAPI_HOST", default="0.0.0.0"
    )
    log_to_console = config.get(
        "LOG_TO_CONSOLE", default=True
    )
    log_dir = config.get(
        "LOGING_DIR", default=default_root_dir()
    )
    log_filename = config.get(
        "LOG_FILENAME", default=log_filename_default
    )

    return ApiConfig(
        logging_level=logging_level,
        cache_size=cache_size,
        cache_timeout=cache_timeout,
        server_port=server_port,
        server_host=server_host,
        log_to_console=str_to_bool(log_to_console),
        log_dir=log_dir,
        log_filename=log_filename
    )



import logging
from collections.abc import Callable
from functools import wraps

from fastapi import Request

from telicent_lbapi.context import ModelContext

__license__ = """
Copyright (c) Telicent Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

log = logging.getLogger(__name__)


def set_model_class(model_class: type) -> Callable:
    """
    Allows to temporary change the model class endpoints operate on,
    on response API's middleware resets the model class to the one
    rest-service was initialised with
    :param model_class:
    :return:
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(request: Request, *args, **kwargs):
            log.info(f"Model class changed to: {model_class} via decorator.")
            ModelContext.set_model_class(model_class)
            return await func(request, *args, **kwargs)
        return wrapper

    return decorator



import logging

from fastapi import APIRouter, HTTPException, Request
from pydantic import ValidationError

from telicent_lbapi.utils.cache import get_cached_labels
from telicent_lbapi.utils.helpers import log_extra

__license__ = """
Copyright (c) Telicent Ltd.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

log = logging.getLogger(__name__)
router = APIRouter()


@router.post("")
async def ingest_data_header(request: Request):
    try:
        data_header = await request.json()
    except Exception as e:
        log.exception("Error processing request.", exc_info=e)
        raise HTTPException(status_code=400, detail="Error processing request.") from e

    try:
        security_label = await get_cached_labels(data_header)
    except ValidationError as e:
        log.error("Validation error occurred.", exc_info=e)
        raise HTTPException(status_code=400, detail="Validation error occurred.") from e
    except ValueError as e:
        log.error("Failed to build security labels.", exc_info=e)
        raise HTTPException(status_code=400, detail=str(e)) from e
    except Exception as e:
        log.error("An unexpected error occurred.", exc_info=e)
        raise HTTPException(status_code=500, detail="Internal server error") from e

    log.info("Request - ok", extra=log_extra(request))
    return {"status": "success", "security_label": security_label}


@router.get("/status")
async def get_status():
    return {"status": "OK"}



import logging
from contextlib import asynccontextmanager

import uvicorn
from fastapi import APIRouter, FastAPI
from fastapi.routing import APIRoute
from starlette.middleware import Middleware
from starlette_context.middleware import RawContextMiddleware
from starlette_context.plugins import RequestIdPlugin

from telicent_lbapi.api.v1 import router as api_router
from telicent_lbapi.context import ModelContext
from telicent_lbapi.core.api_config.config import load_config
from telicent_lbapi.core.api_config.log_config import configure_logging
from telicent_lbapi.core.middleware.context_middleware import ModelContextMiddleware

__license__ = """
Copyright (c) Telicent Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

log = logging.getLogger(__name__)


def create_app(model_class: type, custom_router: APIRouter = None):
    config = load_config()
    configure_logging(config)

    middleware = [
        Middleware(RawContextMiddleware, plugins=(RequestIdPlugin(),)),
        Middleware(ModelContextMiddleware, default_model_class=model_class)
    ]

    app = FastAPI(middleware=middleware)

    app.include_router(api_router, prefix="/api/v1")

    if custom_router:
        app.include_router(custom_router, prefix="/custom")

    ModelContext.set_model_class(model_class)

    @asynccontextmanager
    async def lifespan(ap: FastAPI):
        log.info(f"Starting {model_class.__name__} REST API service")
        log.info(f"Configuration: {config}")
        for route in ap.routes:
            if isinstance(route, APIRoute):
                log.info(f"Path: {route.path}, Name: {route.name}, Methods: {route.methods}")
        yield
        log.info(f"Shutting down {model_class.__name__} REST API service")

    app.router.lifespan_context = lifespan

    return app, config


def run_api_service(model_class: type, custom_router: APIRouter = None):
    app, config = create_app(model_class, custom_router)
    uvicorn.run(app, host=config.server_host, port=config.server_port, log_config=None)
